/***********************************************************************/
/*  gfx.h - Video functions     																			 */
/***********************************************************************/

#include <conio.h>
#include <string.h>
#include <i86.h>
#include <fstream.h>
#include <iostream.h>
#include <stdlib.h>

#ifndef _hiteck_gfx_h_
#define _hiteck_gfx_h_

//#include "structs.h"
#include "palette.h"
#include "pcx.h"

typedef struct
{
	long x1, y1, x2, y2;
} cliptype;


#ifdef hicolor
	#ifndef pal256
		typedef struct
		{
			short r;
			short g;
			short b;
		} bitmapdata;
	#else
		typedef unsigned char    bitmapdata;
	#endif
#else
	typedef unsigned char    bitmapdata;
#endif

typedef struct
{
	unsigned int xsize;
	unsigned int ysize;
	unsigned short mode;
	bitmapdata trans;
	bitmapdata *data;
} bitmaptype;

typedef struct 
{
	bitmaptype *data;
	short frames;
} sprite;


char busy()
{
	static char pos;
	char busy[4] = "-\|/";
	pos = (pos+1) % 4;
	return busy[pos];
}

void error(char *message)
{
	cerr << message << endl;
	exit(1);
}

typedef signed          long    int     slong;
typedef unsigned        long    int     ulong;
typedef signed          short   int     sword;
typedef unsigned        short   int     uword;
typedef signed          char            sbyte;
typedef unsigned        char            ubyte;


typedef struct
{
	unsigned short int Segment;
	unsigned short int Selector;
} RealPointer;
                

typedef struct
{
	char 						signature[4];
	unsigned short 	version;
	unsigned long 	 OEMnamepointer;
	unsigned long 	 capabilities;
	unsigned long 	 modelistpointer;
	unsigned short 	memoryblocks;
	char 						reserved[493];
} VesaInfo;

typedef struct
{
	uword           ModeAttributes;
	ubyte           WindowAAttributes;
	ubyte           WindowBAttributes;
	uword           WindowGranularity;
	uword           WindowSize;
	uword           StartSegmentWindowA;
	uword           StartSegmentWindowB;
	void            (*WindowPositioningFunction)(slong page);
	uword           BytesPerScanLine;

	uword           PixelWidth;
	uword           PixelHeight;
	ubyte           CharacterCellPixelWidth;
	ubyte           CharacterCellPixelHeight;
	ubyte           NumberOfMemoryPlanes;
	ubyte           BitsPerPixel;
	ubyte           NumberOfBanks;
	ubyte           MemoryModelType;
	ubyte           SizeOfBank;
	ubyte           NumberOfImagePages;
	ubyte           Reserved1;
                
	ubyte           RedMaskSize;
	ubyte           RedFieldPosition;
	ubyte           GreenMaskSize;
	ubyte           GreenFieldPosition;
	ubyte           BlueMaskSize;
	ubyte           BlueFieldPosition;
	ubyte           ReservedMaskSize;
	ubyte           ReservedFieldPosition;
	ubyte           DirectColourModeInfo;
	
	ulong        	 PhysBasePtr;
	ulong        	 OffScreenMemOffset;
	uword 					OffScreenMemSize;

	ubyte           Reserved2[206];
} VesaModeData;
                                                                                                        


typedef struct RMREGS
{
	ulong		edi;
	ulong		esi;
	ulong		ebp;
	ulong		reserved;
	ulong		ebx;
	ulong		edx;
	ulong		ecx;
	ulong		eax;
	uword		flags;
	uword		es,ds,fs,gs,ip,cs,sp,ss;
} RMREGS;


RealPointer				vesa_info_rp;
RealPointer				vesa_mode_data_rp;


VesaInfo			 far	*	vbe_info		 	= NULL;
VesaModeData	 far	*	vesa_mode_data = NULL;
unsigned char       *  linear;
cliptype 			defclip;


short useLFB=0;
slong					vesa_granularity;
slong					vesa_page;
slong					vesa_width;
slong					vesa_height;
slong					vesa_bits_per_pixel;
slong					vesa_bytes_per_pixel;



typedef struct {
  unsigned long handle;
  unsigned long size;
  unsigned long address;
} __dpmi_meminfo;


/*
typedef unsigned char  uchar;
typedef unsigned short ushort;
typedef unsigned long  ulong;
typedef unsigned char  BYTE;
*/

int __dpmi_free_physical_address_mapping(__dpmi_meminfo *_info)/* DPMI 0.9 AX=0801 */
{
    union REGS r;
    r.w.ax = 0x0801;
    int386(0x31,&r,&r);
    return r.x.cflag;
}

int __dpmi_free_ldt_descriptor(int _descriptor)
/* DPMI 0.9 AX=0001 */
{
    union REGS r;
    r.w.ax = 1;
    r.w.bx = _descriptor;
    int386(0x31,&r,&r);
    return r.x.cflag;
}


int __dpmi_allocate_ldt_descriptors(int _count)
/* DPMI 0.9 AX=0000 */
{
    union REGS r;
    r.w.ax = 0;
    r.w.cx = _count;
    int386(0x31,&r,&r);
    if (r.x.cflag) cerr << "PM allocSelector failed";
    if (r.x.cflag) exit(1);
    return r.w.ax;
}


int __dpmi_set_descriptor_access_rights(int _selector, int _rights)
/* DPMI 0.9 AX=0009 */
{
    union REGS r;
    r.w.ax = 0x0009;
    r.w.bx = _selector;
    r.w.cx = _rights;
    int386(0x31,&r,&r);
    return r.x.cflag;
}


int __dpmi_physical_address_mapping(__dpmi_meminfo *_info)/* DPMI 0.9 AX=0800 */
{
    union REGS r;
    r.w.ax = 0x800;
    r.w.bx = (short) (_info->address >> 16);
    r.w.cx = (short) (_info->address & 0xFFFF);
    r.w.si = (short) (_info->size  >>16 );
    r.w.di = (short) (_info->size   & 0xFFFF);
    int386(0x31,&r,&r);
    _info->handle = ((ulong)r.w.bx << 16) + (ulong)r.w.cx;
    return r.x.cflag;
}


int __dpmi_set_segment_base_address(int _selector, unsigned long _address)
/* DPMI 0.9 AX=0007 */
{
    union REGS r;
    r.w.ax = 7;
    r.w.bx = _selector;
    r.w.cx = _address >> 16;
    r.w.dx = _address & 0xFFFF;
    int386(0x31,&r,&r);
    return r.x.cflag;
}


int __dpmi_set_segment_limit(int _selector, unsigned long _limit)
/* DPMI 0.9 AX=0008 */
{
    union REGS r;
    r.w.ax = 8;
    r.w.bx = _selector;
    r.w.cx = _limit >> 16;
    r.w.dx = _limit & 0xFFFF;
    int386(0x31,&r,&r);
    return r.x.cflag;
}


int __dpmi_get_segment_base_address(int _selector, unsigned long *_addr)
/* DPMI 0.9 AX=0006 */
{
    union REGS r;
    r.w.ax = 0x0006;
    r.w.bx = _selector;
    int386(0x31,&r,&r);
    *_addr = ((long)r.w.cx<<16)+r.w.dx;
    return r.x.cflag;
}

unsigned long VESA_linAddr = 0;
unsigned short VESA_selAddr = 0;
short VESA_selector=0;

void freeLFB()
{
	__dpmi_meminfo info;
	info.address = VESA_linAddr;
	__dpmi_free_physical_address_mapping(&info);
	__dpmi_free_ldt_descriptor(VESA_selAddr);
}



void far *getLFBaddress(long physaddr,short *selector)
{
    __dpmi_meminfo info;
    if (VESA_linAddr)
    {
			info.address = VESA_linAddr;
			__dpmi_free_physical_address_mapping(&info);
 			VESA_linAddr = 0;
    }
    if (VESA_selAddr)
    {
       __dpmi_free_ldt_descriptor(VESA_selAddr);
		 	VESA_selAddr = 0;
    }
    VESA_selAddr = __dpmi_allocate_ldt_descriptors(1);
    if (__dpmi_set_descriptor_access_rights(VESA_selAddr,0x8092)) 
    {
    	cerr << "PM rights failed" << endl;
    	exit(1);
    }
    info.address = physaddr;
    info.size    = 4096L*1024L-1L;
    if (!__dpmi_physical_address_mapping(&info))
    {
       VESA_linAddr = info.handle;
       __dpmi_set_segment_base_address(VESA_selAddr,VESA_linAddr);
       __dpmi_set_segment_limit(VESA_selAddr,4096L*1024L-1L);
       __dpmi_get_segment_base_address(VESA_selAddr,&VESA_linAddr);
		   useLFB = 1;
       linear = (unsigned char*)VESA_linAddr;
       *selector = (short)VESA_selAddr;
       return MK_FP(VESA_selAddr,0);
//			 return (ulong*)VESA_linAddr; //<<= Pointer to Video
    }
		cerr << "Cannot remap address!" << endl;
    *selector   = 0;
    useLFB = 0;
    return NULL;
}



void DWORDCopy(unsigned char *Source, unsigned char *Destination, int Length);
#pragma aux DWORDCopy = " rep movsd " parm [ESI] [EDI] [ECX]\
modify [ESI EDI ECX];

/*
void getLFBaddress(unsigned long phys, long memory);
#pragma aux getLFBaddress =		\
    "mov    cx,bx										" \
    "shr    ebx,16									 " \
    "xor    edi,edi 								 " \
    "mov    ax,0800h                 " \
    "int    31h										  " \
    "shl    ebx,16							     " \
    "mov    bx,cx									  " \
    "mov    [linear],ebx 				    " \
    parm [ebx] [si]\
    modify [ax bx cx ebx esi edi si];
*/



short getbit(long byte, short bit)
{
	switch(bit)
	{
		case 1: if (byte & 128) return 1;
		case 2: if (byte &  64) return 1;
		case 3: if (byte &  32) return 1;
		case 4: if (byte &  16) return 1;
		case 5: if (byte &   8) return 1;
		case 6: if (byte &   4) return 1;
		case 7: if (byte &   2) return 1;
		case 8: if (byte &   1) return 1;
	}
	return 0;
}



void setmode(int mode)
{
    union REGS r;
    r.x.eax = 0x4F02;
    r.x.ebx = mode;

    int386(0x10,&r,&r);
}

inline void pset (int x, int y, int col)
{
    char *screen;
    
    screen=(char *) 0xa0000;                   
    screen[y*320+x] = col;
}



#ifdef hires256
inline void vpset (unsigned char *vbuff, int x, int y, int col)
{
    vbuff[y*vesa_mode_data->BytesPerScanLine+x] = col;
}
#else
inline void vpset (unsigned char *vbuff, int x, int y, int col)
{
    vbuff[y*320+x] = col;
}
#endif




#ifdef hires256
inline unsigned char vpoint(unsigned char *vbuff, int x, int y)
{
    return vbuff[y*vesa_mode_data->BytesPerScanLine+x];
}
#else
inline unsigned char vpoint(unsigned char *vbuff, int x, int y)
{
    return vbuff[y*320+x];
}
#endif



inline unsigned char point(int x, int y)
{
   char *screen;
   screen=(char *) 0xa0000;                   
   return screen[y*320+x];
}

void cls()
{
    int buffsize;
    char *screen;
    
    screen=(char *) 0xa0000;                   
    for (buffsize=0;buffsize<64000;buffsize++) screen[buffsize]=0;
}

void vcls(unsigned char *vbuff, slong swidth,slong sheight,slong bpp)
{
    unsigned int *point=(unsigned int *)vbuff, count,
        size = (unsigned int)(swidth*sheight*bpp)/4;

    for (count=0;count<size;count++)
        point[count]=0;
}

void makectable(char *pal, unsigned char *ctable)
{
  ctable = (unsigned char *)malloc((64*64*64) * sizeof(unsigned char));
  if (ctable == NULL) error(" * Out of memory: hicolor table");
	generatecolortable(ctable, pal);
}


inline void smooth(unsigned char *what, short xsize, short ysize, short bpp, int iter, int rate)
{
	if (bpp==1)
	{
	  unsigned long size = (long)(xsize*ysize) - xsize - 1;
	  
	  for (int i=0;i<iter;i++)
	  	for (unsigned long ofs=xsize+1;ofs<size;ofs++)
	  	{
	  		what[ofs] = (what[ofs-1] + what[ofs+1] + what[ofs-xsize] + what[ofs+xsize]) / 4;
	  	}
	}    
	
	
	if (bpp==4)
	{
		xsize *= 4;
	  unsigned long size = (long)(xsize*ysize) - xsize - 4;
	  
	  for (long i=0;i<iter;i++)
	  	for (unsigned long ofs=xsize+4;ofs<size;ofs++)
	  		what[ofs] = (unsigned char)((what[ofs-4] + what[ofs+4] + what[ofs-xsize] + what[ofs+xsize]) >> 2);
	}
}


inline void copytopage(unsigned char *array, unsigned char *vbuff)
{
	int buffsize=64000;
  memcpy(&vbuff[0],&array[0],buffsize);
}

void set666pal(char *pal)
{
	int count=0;
	for (int i=0;i<768;i++) pal[i] = 0;
	
	for (int r=0;r<6;r++)
			for (int g=0;g<6;g++)
					for (int b=0;b<6;b++)
					{
						pal[(count)*3+0] = (r*51)/4;
						pal[(count)*3+1] = (g*51)/4;
						pal[(count++)*3+2] = (b*51)/4;
					}
}

/*
#ifdef pal256
	#ifdef hires256
		inline void vpseth (unsigned char *vbuff, int x, int y, rgbtype col, unsigned char *ctable)
		{
		    vbuff[y*vesa_mode_data->BytesPerScanLine+x] = ctable[(col.r*4096)+(col.g*64)+col.b];
		}
		inline void vpseth2 (unsigned char *vbuff, int x, int y, rgbdelta col, unsigned char *ctable)
		{
		    vbuff[y*vesa_mode_data->BytesPerScanLine+x] = ctable[((col.r/261344)*4096)+((col.g/261344)*64)+(col.b/261344)];
		}
	#else
		inline void vpseth (unsigned char *vbuff, int x, int y, rgbtype col, unsigned char *ctable)
		{
		    vbuff[y*320+x] = ctable[(col.r*4096)+(col.g*64)+col.b];
		}
		inline void vpseth2 (unsigned char *vbuff, int x, int y, rgbdelta col, unsigned char *ctable)
		{
	    vbuff[y*320+x] = ctable[((col.r/261344)*4096)+((col.g/261344)*64)+(col.b/261344)];
		}
	#endif
#else
	inline void vpseth (unsigned char *vbuff, int x, int y, rgbtype col)
	{
			long off=y*vesa_mode_data->BytesPerScanLine+(x*vesa_bytes_per_pixel);
	  	vbuff[off] = col.r;
		  vbuff[off+1] = col.g;
		  vbuff[off+2] = col.b;
	}
	
	inline void vpseth2 (unsigned char *vbuff, int x, int y, rgbdelta col)
	{
			long off=y*vesa_mode_data->BytesPerScanLine+(x*vesa_bytes_per_pixel);
	  	vbuff[off] = col.r/65536;
		  vbuff[off+1] = col.g/65536;
		  vbuff[off+2] = col.b/65536;
	}
#endif

#ifdef hicolor
	#ifdef pal256
		#ifdef hires256
			inline void vpsett (unsigned char *vbuff, int x, int y, char col, unsigned char *ttable)
			{
				vbuff[y*vesa_mode_data->BytesPerScanLine+x]=ttable[vbuff[y*vesa_mode_data->BytesPerScanLine+x]*256+col];
			}
		#else
			inline void vpsett (unsigned char *vbuff, int x, int y, char col, unsigned char *ttable)
			{
				vbuff[y*320+x]=ttable[vbuff[y*320+x]*256+col];
			}
		#endif
	#else
		inline void vpsett (unsigned char *vbuff, int x, int y, rgbtype col)
		{
			long off=y*vesa_mode_data->BytesPerScanLine+(x*vesa_bytes_per_pixel);

	    vbuff[off] = (vbuff[off]+col.b/2);
	    vbuff[off+1] = (vbuff[off+1]+col.g/2);
	    vbuff[off+2] = (vbuff[off+2]+col.r/2);
		}
		inline void vpsett2 (unsigned char *vbuff, int x, int y, rgbdelta col)
		{
			long off=y*vesa_mode_data->BytesPerScanLine+(x*vesa_bytes_per_pixel);

	    vbuff[off] = (vbuff[off]+col.b/65536/2);
	    vbuff[off+1] = (vbuff[off+1]+col.g/65536/2);
	    vbuff[off+2] = (vbuff[off+2]+col.r/65536/2);
		}
	#endif
#else
	inline void vpsett (unsigned char *vbuff, int x, int y, char col, unsigned char *ttable)
	{
		vbuff[y*320+x]=ttable[vbuff[y*320+x]*256+col];
	}
#endif
*/










static void far * allocate_dos_memory(RealPointer * rp,ulong bytes_to_allocate)
{
	void far *ptr=NULL;
	union REGS regs;

	bytes_to_allocate = ((bytes_to_allocate + 15) & 0xfffffff0);
	memset(&regs,0,sizeof(regs));
	regs.w.ax = 0x100;
	regs.w.bx = (uword)(bytes_to_allocate >> 4);
	int386(0x31,&regs,&regs);
	if(regs.x.cflag == 0)
	{
		rp->Segment = regs.w.ax;
		rp->Selector = regs.w.dx;
		ptr = MK_FP(regs.w.dx,0);
	}
	return(ptr);
}


static void free_dos_memory(RealPointer * rp)
{
	union REGS regs;

	regs.w.ax = 0x101;
	regs.w.dx = rp->Selector;
	int386(0x31,&regs,&regs);
}

signed int VBE_getinfo()
{
	union REGS	regs;
	struct SREGS	sregs;
	RMREGS rmregs;

	if((vbe_info = (VesaInfo far *)allocate_dos_memory(&vesa_info_rp,sizeof(VesaInfo))) != NULL)
	{
		memset(&rmregs,0,sizeof(rmregs));
		memset(&regs,0,sizeof(regs));
		memset(&sregs,0,sizeof(sregs));
		segread(&sregs);
		rmregs.eax = 0x4f00;
		rmregs.es = vesa_info_rp.Segment;
		rmregs.ds = vesa_info_rp.Segment;
		regs.x.eax = 0x300;
		regs.x.ebx = 0x10;
		regs.x.ecx = 0;
		sregs.es = FP_SEG(&rmregs);
		regs.x.edi = FP_OFF(&rmregs);
		int386x(0x31,&regs,&regs,&sregs);
		if(rmregs.eax == 0x4f)
		{
			return 1;
		}
	}
	return 0;
}



signed int VBE_getmodeinfo(uword vmode)
{
	union REGS regs;
	struct SREGS sregs;
	RMREGS rmregs;

	if((vesa_mode_data = (VesaModeData far *)allocate_dos_memory(&vesa_mode_data_rp,sizeof(VesaModeData))) != NULL)
	{
		memset(&rmregs,0,sizeof(rmregs));
		rmregs.es = vesa_mode_data_rp.Segment;
		rmregs.ds = vesa_mode_data_rp.Segment;
		rmregs.edi = 0;
		rmregs.eax = 0x4f01;
		rmregs.ecx = vmode;
		memset(&regs,0,sizeof(regs));
		memset(&sregs,0,sizeof(sregs));
		segread(&sregs);
		regs.x.eax = 0x300;
		regs.x.ebx = 0x10;
		regs.x.edi = (ulong)&rmregs;
		int386x(0x31,&regs,&regs,&sregs);
		if(regs.h.al == 0)
		{
			vesa_granularity	= vesa_mode_data->WindowGranularity;
			vesa_width			= vesa_mode_data->PixelWidth;
			vesa_height			= vesa_mode_data->PixelHeight;
			vesa_bits_per_pixel	= vesa_mode_data->BitsPerPixel;
			if (vesa_bits_per_pixel==1)  vesa_bytes_per_pixel = 1/8;
			if (vesa_bits_per_pixel==4)  vesa_bytes_per_pixel = 0.5;
			if (vesa_bits_per_pixel==8)  vesa_bytes_per_pixel = 1;
			if (vesa_bits_per_pixel==15)  vesa_bytes_per_pixel = 2;
			if (vesa_bits_per_pixel==16)  vesa_bytes_per_pixel = 2;
			if (vesa_bits_per_pixel==24)  vesa_bytes_per_pixel = 3;
			if (vesa_bits_per_pixel==32)  vesa_bytes_per_pixel = 4;
	    defclip.x1 = 0;
	    defclip.y1 = 0;
	    defclip.x2 = vesa_width;
	    defclip.y2 = vesa_height;
			return 1;
		}
	}
	return 0;
}


void vesa_setpage(slong vpage)
{
	union REGS regs;
	vesa_page = vpage;
	regs.w.ax = 0x4f05;
	regs.w.bx = 0;
	regs.w.dx = vpage;
	int386(0x10,&regs,&regs);
}


void flippage(ubyte *source,slong swidth,slong sheight,slong bpp)
{
		if (useLFB==1)
		{
			slong scr_size = (swidth * (sheight) * bpp);
			DWORDCopy(source,(unsigned char *)linear,scr_size/4);
		} 
		else 
		{
			void	*dest;
			slong page = 0;
			slong scr_size	= (swidth * sheight * bpp);

			dest = (void *)source;
//		DWORDCopy(system->display.vbuff,(unsigned char *)linear,scr_size/4);

			while(scr_size > 0)
			{
				vesa_setpage(page);
				if (scr_size>65535)
//					memcpy((void *)0xa0000,dest,65536);
      		DWORDCopy((unsigned char*)dest,(unsigned char *)0xa0000,0x4000);
				else
					memcpy((void *)0xa0000,dest,scr_size);
				scr_size -= 65536;
				dest = (void *)((ulong)dest + 65536);
				page++;
			}
		}
}

unsigned int gfx_findmode(int xres, int yres, int bpp)
{
	for (int mode=0x100;mode<0x400;mode++)
	{
		if (VBE_getmodeinfo(mode)) 
		{
			if ((vesa_width==xres) && 
					(vesa_height==yres) && 
					(vesa_bits_per_pixel==bpp))
					return mode;
		}
	}
	return -1;
}














void getimg(unsigned char *vbuff, long x, long y, int xsize, int ysize,
			      bitmapdata *data, cliptype clip)
{
	unsigned int *point4 = (unsigned int*)vbuff;
	unsigned short *point2 = (unsigned short*)vbuff;
	unsigned int *imgpoint4 = (unsigned int*)data;
	unsigned short *imgpoint2 = (unsigned short*)data;
  unsigned long xx, yy;

  if (((x+xsize)<clip.x1) || ((y+ysize)<clip.y1)) return;
  if ((x>clip.x2) || (y>clip.y2)) return;
  
  long xoffset=0, yoffset=0;
	long tempxsize=xsize;
	
	if ((x>vesa_width-1) || (y>vesa_height-1) || ((x+xsize)<0) || ((y+ysize)<0)) return;
	
  if (x<clip.x1) { xoffset= -(x-clip.x1); xsize-=xoffset; x=clip.x1; }
  if (y<clip.y1) { yoffset= -(y-clip.y1); ysize-=yoffset; y=clip.y1; }
  
	if ((x+xsize)>(clip.x2)) xsize = clip.x2-x;
	if ((y+ysize)>(clip.y2)) ysize = clip.y2-y;
  x*=vesa_bytes_per_pixel;
  
	unsigned long ofs, ofs2, ofs3;
	if (!(xsize>0) && (ysize>0)) return;
	  
	ofs = y*vesa_mode_data->BytesPerScanLine+x;
	ofs2 = (yoffset)*tempxsize + xoffset;
	
	#if (defined(hicolor) && !defined(pal256))
		if (vesa_bytes_per_pixel==4) ofs /= 4;
		if (vesa_bytes_per_pixel==2) ofs /= 2;
  #endif
	
	if (vesa_bytes_per_pixel==2)
	  for (yy=y;yy<y+ysize;yy++)
	  {
			ofs3=ofs;
		  for (xx=ofs2;xx<ofs2+xsize;xx++)
		  {
  			imgpoint2[xx] = point2[ofs3++];
		  }
			ofs+=(vesa_mode_data->BytesPerScanLine)/2;
			ofs2+=tempxsize;	
	  }
	else	  
  for (yy=y;yy<y+ysize;yy++)
  {
		#if (defined(hicolor) && !defined(pal256))
			ofs3=ofs;
		  for (xx=ofs2;xx<ofs2+xsize;xx++)
		  {
   			imgpoint4[xx] = point4[ofs3++];
		  }
			ofs+=vesa_mode_data->BytesPerScanLine/4;	
			ofs2+=tempxsize;	
    #else
			ofs3=ofs;
		  for (xx=ofs2;xx<ofs2+xsize;xx++)
		  {
		    data[xx] = vbuff[ofs3++];
		  }
			ofs+=vesa_mode_data->BytesPerScanLine;
			ofs2+=tempxsize;
    #endif
  }
}





#if (defined(hicolor) && !defined(pal256))
signed int load_bitmap (bitmaptype *bitmap,
												char *filename, int xsize, int ysize,
												int mode,
												int transR, int transG, int transB, float gamma)
#else
signed int load_bitmap (bitmaptype *bitmap,
												char *filename, int xsize, int ysize,
												int mode,
												bitmapdata trans, float gamma, unsigned char *ctable)
#endif												
{
  bitmap->data = (bitmapdata *)malloc(sizeof(bitmapdata) * ((xsize*ysize)+1));
	bitmap->xsize = xsize;
	bitmap->ysize = ysize;
	bitmap->mode = mode;
	
	#if (defined(hicolor) && !defined(pal256))
		bitmapdata temp;
		temp.r = transR;
		temp.g = transG;
		temp.b = transB;
		bitmap->trans = temp;
	#else
		bitmap->trans = trans;
	#endif
	
  char temppal[768];
  unsigned char *tempdata = (unsigned char *)malloc(sizeof(unsigned char) * ((xsize*ysize)+1));
  
  if (bitmap->data == NULL) error(" * Out of memory: bitmap data");
  if (tempdata == NULL) error(" * Out of memory: temporary bitmap data");
  
  memset(bitmap->data, 0, sizeof(bitmapdata)*((xsize*ysize)+1));
  memset(tempdata, 0, sizeof(unsigned char)*((xsize*ysize)+1));
  
  loadpcx(filename, xsize, ysize, tempdata);
  loadpal(filename, &temppal[0]);
  unsigned char r,g,b;
  
	for (int y=0;y<ysize;y++)
		for (int x=0;x<xsize;x++)
		{
			#ifdef hicolor
			  r = (float)temppal[tempdata[y*xsize+x]*3+0]*gamma;
				g = (float)temppal[tempdata[y*xsize+x]*3+1]*gamma;
				b = (float)temppal[tempdata[y*xsize+x]*3+2]*gamma;
				#ifdef pal256
					if (r>63) r=63;
					if (g>63) g=63;
					if (b>63) b=63;
					bitmap->data[y*xsize+x] = ctable[(r*4096)+(g*64)+b];
				#else																				
					if (r>255) r=255;
					if (g>255) g=255;
					if (b>255) b=255;
					bitmap->data[y*xsize+x].r = r;
					bitmap->data[y*xsize+x].g = g;
					bitmap->data[y*xsize+x].b = b;
				#endif				
			#else
				bitmap->data[y*xsize+x] = tempdata[y*xsize+x];
			#endif				
	}
	free (tempdata);
	return -1;
}


















//////////////////////////////////////////////
// RAW BitBlt's													 	//
//////////////////////////////////////////////

void putraw(unsigned char *vbuff, long x, long y, int xsize, int ysize,
			      bitmapdata *data, short mode, unsigned char *ttable,
			      bitmapdata trans, cliptype clip)
{
	unsigned int *point = (unsigned int*)vbuff;
  unsigned long xx, yy;

  if (((x+xsize)<clip.x1) || ((y+ysize)<clip.y1)) return;
  if ((x>clip.x2) || (y>clip.y2)) return;
  
  long xoffset=0, yoffset=0;
	long tempxsize=xsize;
	
	if ((x>vesa_width-1) || (y>vesa_height-1) || ((x+xsize)<0) || ((y+ysize)<0)) return;
	
  if (x<clip.x1) { xoffset= -(x-clip.x1); xsize-=xoffset; x=clip.x1; }
  if (y<clip.y1) { yoffset= -(y-clip.y1); ysize-=yoffset; y=clip.y1; }
  
	if ((x+xsize)>(clip.x2)) xsize = clip.x2-x;
	if ((y+ysize)>(clip.y2)) ysize = clip.y2-y;
  x*=vesa_bytes_per_pixel;
  
	unsigned long ofs, ofs2, ofs3;
	if (!(xsize>0) && (ysize>0)) return;
	  
	ofs = y*vesa_mode_data->BytesPerScanLine+x;
	ofs2 = (yoffset)*tempxsize + xoffset;
	
	#if (defined(hicolor) && !defined(pal256))
		if (mode < 2) ofs /= 4;
  #endif
	
	switch (mode)
	{
	case 0:															//plain
	  for (yy=y;yy<y+ysize;yy++)
	  {
			#if (defined(hicolor) && !defined(pal256))
				ofs3=ofs;
			  for (xx=ofs2;xx<ofs2+xsize;xx++)
			  {
//			    vbuff[ofs3] = data[xx].b;
//			    vbuff[ofs3+1] = data[xx].g;
//			    vbuff[ofs3+2] = data[xx].r;
//			    ofs3+=vesa_bytes_per_pixel;
    			point[ofs3++] = (data[xx].b) + (data[xx].g<<8) + (data[xx].r<<16);
			  }
				ofs+=vesa_mode_data->BytesPerScanLine/4;	
				ofs2+=tempxsize;	
	    #else
				ofs3=ofs;
			  for (xx=ofs2;xx<ofs2+xsize;xx++)
			  {
			    vbuff[ofs3] = data[xx];
			    ofs3++;
			  }
				ofs+=vesa_mode_data->BytesPerScanLine;
				ofs2+=tempxsize;
	    #endif
	  }
  break;

	case 1:															//transparent color
	  for (yy=y;yy<y+ysize;yy++)
	  {
			#if (defined(hicolor) && !defined(pal256))
				ofs3=ofs;
			  for (xx=ofs2;xx<ofs2+xsize;xx++)
			  {
			    if ((data[xx].r!=trans.r) && 
			    		(data[xx].g!=trans.g) && 
			    		(data[xx].b!=trans.b))
			    {
//				    vbuff[ofs3] = data[xx].b;
//				    vbuff[ofs3+1] = data[xx].g;
//				    vbuff[ofs3+2] = data[xx].r;
      			point[ofs3] = (data[xx].b) + (data[xx].g<<8) + (data[xx].r<<16);
			    }
			    //ofs3+=vesa_bytes_per_pixel;
			    ofs3++;
			  }
				ofs+=vesa_mode_data->BytesPerScanLine/4;
				ofs2+=tempxsize;
	    #else
				ofs3=ofs;
			  for (xx=ofs2;xx<ofs2+xsize;xx++)
			  {
			  	if (data[xx]!=trans) vbuff[ofs3] = data[xx];
			    ofs3++;
			  }
				ofs+=vesa_mode_data->BytesPerScanLine;
				ofs2+=tempxsize;
	    #endif
	  }
	break;

	case 2:																		//translucent
	  for (yy=y;yy<y+ysize;yy++)
	  {
			#if (defined(hicolor) && !defined(pal256))
				ofs3=ofs;
			  for (xx=ofs2;xx<ofs2+xsize;xx++)
			  {
			    vbuff[ofs3] = (vbuff[ofs3]+data[xx].b)/2;
			    vbuff[ofs3+1] = (vbuff[ofs3+1]+data[xx].g)/2;
			    vbuff[ofs3+2] = (vbuff[ofs3+2]+data[xx].r)/2;
			    ofs3+=vesa_bytes_per_pixel;
			  }
				ofs+=vesa_mode_data->BytesPerScanLine;	
				ofs2+=tempxsize;	
	    #else
				ofs3=ofs;
			  for (xx=ofs2;xx<ofs2+xsize;xx++)
			  {
			    vbuff[ofs3] = ttable[vbuff[ofs3]*256+data[xx]];
			    ofs3++;
			  }
				ofs+=vesa_mode_data->BytesPerScanLine;
				ofs2+=tempxsize;
	    #endif
	  }
  break;


	case 3:																		//translucent with
																						 //transparent color
	  for (yy=y;yy<y+ysize;yy++)
	  {
			#if (defined(hicolor) && !defined(pal256))
				ofs3=ofs;
			  for (xx=ofs2;xx<ofs2+xsize;xx++)
			  {
			    if ((data[xx].r!=trans.r) && 
			    		(data[xx].g!=trans.g) && 
			    		(data[xx].b!=trans.b))
			    {
				    vbuff[ofs3] = (vbuff[ofs3]+data[xx].b)/2;
				    vbuff[ofs3+1] = (vbuff[ofs3+1]+data[xx].g)/2;
				    vbuff[ofs3+2] = (vbuff[ofs3+2]+data[xx].r)/2;
				  }
			    ofs3+=vesa_bytes_per_pixel;
			  }
				ofs+=vesa_mode_data->BytesPerScanLine;	
				ofs2+=tempxsize;	
	    #else
				ofs3=ofs;
			  for (xx=ofs2;xx<ofs2+xsize;xx++)
			  {
			    if (data[xx]!=trans) vbuff[ofs3] = ttable[vbuff[ofs3]*256+data[xx]];
			    ofs3++;
			  }
				ofs+=vesa_mode_data->BytesPerScanLine;
				ofs2+=tempxsize;
	    #endif
	  }
	break;
  }
}








void putscaledraw(unsigned char *vbuff, long x, long y, long x2, long y2,
			            bitmapdata *data, int xsize, int ysize, short mode, unsigned char *ttable,
			            bitmapdata trans, cliptype clip)
{
  unsigned int *point = (unsigned int*)vbuff;
  unsigned long xx, yy, temp;
  if ((x2==x) || (y2==y)) return;

  if ((x2<clip.x1) || (y2<clip.y1)) return;
  if ((x>clip.x2) || (y>clip.y2)) return;
  
  int txsize=xsize;
  int tysize=ysize;
  xsize=x2-x;
  ysize=y2-y;

  unsigned long xoffset=0, yoffset=0;
	unsigned long tempxsize=txsize;
	
	if ((x>vesa_width-1) || (y>vesa_height-1) || ((x+xsize)<0) || ((y+ysize)<0)) return;

	unsigned long txdelta=(txsize)*65536/(x2-x);
	unsigned long tydelta=(tysize)*65536/(y2-y);
	
  if (x<clip.x1) { xoffset= (-(x-clip.x1)*txdelta)/65536; xsize+=(x-clip.x1); x=clip.x1; }
  if (y<clip.y1) { yoffset= (-(y-clip.y1)*tydelta)/65536; ysize+=(y-clip.y1); y=clip.y1; }
  
	if ((x+xsize)>clip.x2) xsize = clip.x2-x;
	if ((y+ysize)>clip.y2) ysize = clip.y2-y;
  x*=vesa_bytes_per_pixel;
  
	unsigned long ofs, ofs2, ofs3;
	unsigned long tx, ty=yoffset*65536;
	
	if (!(xsize>0) && (ysize>0)) return;
	  
	ofs = y*vesa_mode_data->BytesPerScanLine+x;

	#if (defined(hicolor) && !defined(pal256))
		if (mode < 2) ofs /= 4;
  #endif

	ofs2 = (yoffset)*tempxsize + xoffset;
	
	switch (mode)
	{
	case 0:																//plain
	  for (yy=y;yy<y+ysize;yy++)
	  {
			tx=(ofs2+xoffset)*65536;
			#if (defined(hicolor) && !defined(pal256))
				ofs3=ofs;
			  for (xx=ofs2;xx<ofs2+xsize;xx++)
			  {
			  	temp=tx/65536;
//			    vbuff[ofs3] = data[temp].b;
//			    vbuff[ofs3+1] = data[temp].g;
//			    vbuff[ofs3+2] = data[temp].r;
    			point[ofs3++] = (data[temp].b) + (data[temp].g<<8) + (data[temp].r<<16);
//			    ofs3+=vesa_bytes_per_pixel;
			    tx+=txdelta;
			  }
				ofs+=vesa_mode_data->BytesPerScanLine/4;
				ofs2=((ty/65536)*tempxsize);
	    #else
				ofs3=ofs;
			  for (xx=ofs2;xx<ofs2+xsize;xx++)
			  {
			    vbuff[ofs3] = data[tx/65536];
			    tx+=txdelta;
			    ofs3++;
			  }
				ofs+=vesa_mode_data->BytesPerScanLine;
				ofs2=((ty/65536)*tempxsize);
	    #endif
	    ty+=tydelta;
	  }
  break;




	case 1:															//transparent color
	  for (yy=y;yy<y+ysize;yy++)
	  {
			tx=(ofs2+xoffset)*65536;
			#if (defined(hicolor) && !defined(pal256))
				ofs3=ofs;
			  for (xx=ofs2;xx<ofs2+xsize;xx++)
			  {
			  	temp=tx/65536;
			    if ((data[temp].r!=trans.r) && 
			    		(data[temp].g!=trans.g) && 
			    		(data[temp].b!=trans.b))
			    {
       			point[ofs3] = (data[temp].b) + (data[temp].g<<8) + (data[temp].r<<16);
				  }
			    ofs3++;
			    tx+=txdelta;
			  }
				ofs+=vesa_mode_data->BytesPerScanLine/4;	
				ofs2=((ty/65536)*tempxsize);
	    #else
				ofs3=ofs;
			  for (xx=ofs2;xx<ofs2+xsize;xx++)
			  {
			    if (data[tx/65536]!=trans) vbuff[ofs3] = data[tx/65536];
			    tx+=txdelta;
			    ofs3++;
			  }
				ofs+=vesa_mode_data->BytesPerScanLine;
				ofs2=((ty/65536)*tempxsize);
	    #endif
	    ty+=tydelta;
	  }
  break;




	case 2:																//translucent
	  for (yy=y;yy<y+ysize;yy++)
	  {
			tx=(ofs2+xoffset)*65536;
			#if (defined(hicolor) && !defined(pal256))
				ofs3=ofs;
			  for (xx=ofs2;xx<ofs2+xsize;xx++)
			  {
			  	temp=tx/65536;
			    vbuff[ofs3] =   (vbuff[ofs3]   +data[temp].b)/2;
			    vbuff[ofs3+1] = (vbuff[ofs3+1] +data[temp].g)/2;
			    vbuff[ofs3+2] = (vbuff[ofs3+2] +data[temp].r)/2;
			    ofs3+=vesa_bytes_per_pixel;
			    tx+=txdelta;
			  }
				ofs+=vesa_mode_data->BytesPerScanLine;	
				ofs2=((ty/65536)*tempxsize);
	    #else
				ofs3=ofs;
			  for (xx=ofs2;xx<ofs2+xsize;xx++)
			  {
			    vbuff[ofs3] = ttable[vbuff[ofs3]*256+data[tx/65536]];
			    tx+=txdelta;
			    ofs3++;
			  }
				ofs+=vesa_mode_data->BytesPerScanLine;
				ofs2=((ty/65536)*tempxsize);
	    #endif
	    ty+=tydelta;
	  }
  break;




	case 3:																//translucent with
																				 //transparent color
	  for (yy=y;yy<y+ysize;yy++)
	  {
			tx=(ofs2+xoffset)*65536;
			#if (defined(hicolor) && !defined(pal256))
				ofs3=ofs;
			  for (xx=ofs2;xx<ofs2+xsize;xx++)
			  {
			  	temp=tx/65536;
			    if ((data[temp].r!=trans.r) && 
			    		(data[temp].g!=trans.g) && 
			    		(data[temp].b!=trans.b))
			    {
				    vbuff[ofs3] =   (vbuff[ofs3]   +data[temp].b)/2;
				    vbuff[ofs3+1] = (vbuff[ofs3+1] +data[temp].g)/2;
				    vbuff[ofs3+2] = (vbuff[ofs3+2] +data[temp].r)/2;
				  }
			    ofs3+=vesa_bytes_per_pixel;
			    tx+=txdelta;
			  }
				ofs+=vesa_mode_data->BytesPerScanLine;	
				ofs2=((ty/65536)*tempxsize);
	    #else
				ofs3=ofs;
			  for (xx=ofs2;xx<ofs2+xsize;xx++)
			  {
			    if (data[tx/65536]!=trans) vbuff[ofs3] = ttable[vbuff[ofs3]*256+data[tx/65536]];
			    tx+=txdelta;
			    ofs3++;
			  }
				ofs+=vesa_mode_data->BytesPerScanLine;
				ofs2=((ty/65536)*tempxsize);
	    #endif
	    ty+=tydelta;
	  }
  break;

  }
}




// 16-bit

#if (defined(hicolor) && !defined(pal256))

void putraw_16(unsigned short *vbuff, long x, long y, int xsize, int ysize,
			         bitmapdata *data, short mode, unsigned char *ttable,
			         bitmapdata trans, cliptype clip)

{
  unsigned long xx, yy;

  if (((x+xsize)<clip.x1) || ((y+ysize)<clip.y1)) return;
  if ((x>clip.x2) || (y>clip.y2)) return;
  
  long xoffset=0, yoffset=0;
	long tempxsize=xsize;
	
	if ((x>vesa_width-1) || (y>vesa_height-1) || ((x+xsize)<0) || ((y+ysize)<0)) return;
	
  if (x<clip.x1) { xoffset= -(x-clip.x1); xsize-=xoffset; x=clip.x1; }
  if (y<clip.y1) { yoffset= -(y-clip.y1); ysize-=yoffset; y=clip.y1; }
  
	if ((x+xsize)>(clip.x2)) xsize = clip.x2-x;
	if ((y+ysize)>(clip.y2)) ysize = clip.y2-y;
  x*=vesa_bytes_per_pixel;
  
	unsigned long ofs, ofs2, ofs3;
	if (!(xsize>0) && (ysize>0)) return;
	  
	ofs = (y*vesa_mode_data->BytesPerScanLine+x)/2;
	ofs2 = (yoffset)*tempxsize + xoffset;
	
	switch (mode)
	{
	case 0:															//plain
	  for (yy=y;yy<y+ysize;yy++)
	  {
			ofs3=ofs;
		  for (xx=ofs2;xx<ofs2+xsize;xx++)
		  {
		    vbuff[ofs3++] = (data[xx].b) + 
		    								(data[xx].g<<6) + 
		    								(data[xx].r<<11);
		  }
			ofs+=(vesa_mode_data->BytesPerScanLine)/2;
			ofs2+=tempxsize;	
	  }
  break;

	case 1:															//transparent color
	  for (yy=y;yy<y+ysize;yy++)
	  {
			ofs3=ofs;
		  for (xx=ofs2;xx<ofs2+xsize;xx++)
		  {
		    if ((data[xx].r!=trans.r) && 
		    		(data[xx].g!=trans.g) && 
		    		(data[xx].b!=trans.b))
		    {
			    vbuff[ofs3] = (data[xx].b) + 
			    					  	(data[xx].g<<6) + 
			    							(data[xx].r<<11);
		    }
		    ofs3++;
		  }
			ofs+=(vesa_mode_data->BytesPerScanLine)/2;
			ofs2+=tempxsize;
	  }
	break;

	case 2:																		//translucent
	  for (yy=y;yy<y+ysize;yy++)
	  {
			ofs3=ofs;
		  for (xx=ofs2;xx<ofs2+xsize;xx++)
		  {
		  	vbuff[ofs3] = (((vbuff[ofs3] & 31)+data[xx].b)/2) + 
									 	 (((((vbuff[ofs3] & 2016) >> 6)+data[xx].g)/2)<<6) +
								  	  (((((vbuff[ofs3++] & 63488) >> 11)+data[xx].r)/2)<<11);
		  }
			ofs+=(vesa_mode_data->BytesPerScanLine)/2;	
			ofs2+=tempxsize;	
	  }
  break;


	case 3:																		//translucent with
																						 //transparent color
	  for (yy=y;yy<y+ysize;yy++)
	  {
			ofs3=ofs;
		  for (xx=ofs2;xx<ofs2+xsize;xx++)
		  {
		    if ((data[xx].r!=trans.r) && 
		    		(data[xx].g!=trans.g) && 
		    		(data[xx].b!=trans.b))
		    {
			  	vbuff[ofs3] = (((vbuff[ofs3] & 31)+data[xx].b)/2) + 
  									 	 (((((vbuff[ofs3] & 2016) >> 6)+data[xx].g)/2)<<6) +
  								  	  (((((vbuff[ofs3] & 63488) >> 11)+data[xx].r)/2)<<11);
			  }
		    ofs3++;
		  }
			ofs+=(vesa_mode_data->BytesPerScanLine)/2;
			ofs2+=tempxsize;	
	  }
	break;
  }
}









void putscaledraw_16(unsigned short *vbuff, long x, long y, long x2, long y2,
			               bitmapdata *data, int xsize, int ysize, short mode, unsigned char *ttable,
			               bitmapdata trans, cliptype clip)
{
  unsigned long xx, yy, temp;
  if ((x2==x) || (y2==y)) return;

  if ((x2<clip.x1) || (y2<clip.y1)) return;
  if ((x>clip.x2) || (y>clip.y2)) return;
  
  int txsize=xsize;
  int tysize=ysize;
  xsize=x2-x;
  ysize=y2-y;

  unsigned long xoffset=0, yoffset=0;
	unsigned long tempxsize=txsize;
	
	if ((x>vesa_width-1) || (y>vesa_height-1) || ((x+xsize)<0) || ((y+ysize)<0)) return;

	unsigned long txdelta=(txsize)*65536/(x2-x);
	unsigned long tydelta=(tysize)*65536/(y2-y);
	
  if (x<clip.x1) { xoffset= (-(x-clip.x1)*txdelta)/65536; xsize+=(x-clip.x1); x=clip.x1; }
  if (y<clip.y1) { yoffset= (-(y-clip.y1)*tydelta)/65536; ysize+=(y-clip.y1); y=clip.y1; }
  
	if ((x+xsize)>clip.x2) xsize = clip.x2-x;
	if ((y+ysize)>clip.y2) ysize = clip.y2-y;
  x*=vesa_bytes_per_pixel;
  
	unsigned long ofs, ofs2, ofs3;
	unsigned long tx, ty=yoffset*65536;
	
	if (!(xsize>0) && (ysize>0)) return;
	  
	ofs = (y*vesa_mode_data->BytesPerScanLine+x)/2;
	ofs2 = (yoffset)*tempxsize + xoffset;
	
	switch (mode)
	{
	case 0:																//plain
	  for (yy=y;yy<y+ysize;yy++)
	  {
			tx=(ofs2+xoffset)*65536;
			ofs3=ofs;
		  for (xx=ofs2;xx<ofs2+xsize;xx++)
		  {
		  	temp=tx/65536;
		    vbuff[ofs3++] = (data[temp].b) + 
		    								(data[temp].g<<6) + 
		    								(data[temp].r<<11);
		    tx+=txdelta;
		  }
			ofs+=vesa_mode_data->BytesPerScanLine/2;
			ofs2=((ty/65536)*tempxsize);
	    ty+=tydelta;
	  }
  break;




	case 1:															//transparent color
	  for (yy=y;yy<y+ysize;yy++)
	  {
			tx=(ofs2+xoffset)*65536;
			ofs3=ofs;
		  for (xx=ofs2;xx<ofs2+xsize;xx++)
		  {
		  	temp=tx/65536;
		    if ((data[temp].r!=trans.r) && 
		    		(data[temp].g!=trans.g) && 
		    		(data[temp].b!=trans.b))
		    {
			    vbuff[ofs3] = (data[temp].b) + 
			    								(data[temp].g<<6) + 
			    								(data[temp].r<<11);
			  }
		    ofs3++;
		    tx+=txdelta;
		  }
			ofs+=vesa_mode_data->BytesPerScanLine/2;
			ofs2=((ty/65536)*tempxsize);
	    ty+=tydelta;
	  }
  break;




	case 2:																//translucent
	  for (yy=y;yy<y+ysize;yy++)
	  {
			tx=(ofs2+xoffset)*65536;
			ofs3=ofs;
		  for (xx=ofs2;xx<ofs2+xsize;xx++)
		  {
		  	temp=tx/65536;
		  	vbuff[ofs3] = (((vbuff[ofs3] & 31)+data[temp].b)/2) + 
									 	 (((((vbuff[ofs3] & 2016) >> 6)+data[temp].g)/2)<<6) +
								  	  (((((vbuff[ofs3++] & 63488) >> 11)+data[temp].r)/2)<<11);
		    tx+=txdelta;
  	  }
			ofs+=vesa_mode_data->BytesPerScanLine/2;
			ofs2=((ty/65536)*tempxsize);
	    ty+=tydelta;
	  }
  break;




	case 3:																//translucent with
																				 //transparent color
	  for (yy=y;yy<y+ysize;yy++)
	  {
			tx=(ofs2+xoffset)*65536;
			ofs3=ofs;
		  for (xx=ofs2;xx<ofs2+xsize;xx++)
		  {
		  	temp=tx/65536;
		    if ((data[temp].r!=trans.r) && 
		    		(data[temp].g!=trans.g) && 
		    		(data[temp].b!=trans.b))
		    {
			  	vbuff[ofs3] = (((vbuff[ofs3] & 31)+data[temp].b)/2) + 
										 	 (((((vbuff[ofs3] & 2016) >> 6)+data[temp].g)/2)<<6) +
									  	  (((((vbuff[ofs3] & 63488) >> 11)+data[temp].r)/2)<<11);
			  }
		    ofs3++;
		    tx+=txdelta;
		  }
			ofs+=vesa_mode_data->BytesPerScanLine/2;
			ofs2=((ty/65536)*tempxsize);
	    ty+=tydelta;
	  }
  break;

  }
}
#endif





void putimg(unsigned char *vbuff, long x, long y, 
						bitmaptype *bitmap, unsigned char *ttable,
						cliptype clip)
{
#if (defined(hicolor) && !defined(pal256))
	if (vesa_bytes_per_pixel==2)
		putraw_16 ((unsigned short*)vbuff,x,y,
 							bitmap->xsize, bitmap->ysize,
               bitmap->data, bitmap->mode,ttable, bitmap->trans, clip);
	else
#endif
		putraw (vbuff,x,y,
						bitmap->xsize, bitmap->ysize,
            bitmap->data, bitmap->mode,ttable, bitmap->trans, clip);
}

void putscaledimg(unsigned char *vbuff, long x, long y, long x2, long y2,
						      bitmaptype *bitmap, unsigned char *ttable,
						      cliptype clip)
{
#if (defined(hicolor) && !defined(pal256))
	if (vesa_bytes_per_pixel==2)
		putscaledraw_16 ((unsigned short*)vbuff,x,y,x2,y2,bitmap->data,
         						bitmap->xsize, bitmap->ysize,
                     bitmap->mode,ttable, bitmap->trans, clip);
	else
#endif
		putscaledraw (vbuff,x,y,x2,y2,bitmap->data,
      						bitmap->xsize, bitmap->ysize,
                  bitmap->mode,ttable, bitmap->trans, clip);
}

unsigned int rgb(short r, short g, short b)
{
	if (vesa_bytes_per_pixel==2) 
  {
		if (r>255) r=255;
		if (g>255) g=255;
		if (b>255) b=255;
    return ((b/8) + ((g/8)<<6) + ((r/8)<<11));
  }
  
	if (vesa_bytes_per_pixel==4) 
	{
    if (r>255) r=255;
		if (g>255) g=255;
		if (b>255) b=255;
		return (b + (g<<8) + (r<<16));
	}
	if (vesa_bytes_per_pixel==1) return r;
	return 0;
}






// Begin shape drawing methods

void draw_line(unsigned char *vbuff, short x, short y, short x2, short y2, long attr, cliptype clip)
{
	short temp;
	if (x2<x) { temp=x; x=x2; x2=temp; }
	if (y2<y) { temp=y; y=y2; y2=temp; }
	
	if (x<clip.x1) x=clip.x1;
	if (y<clip.y1) y=clip.y1;
	if (x2>clip.x2) x2=clip.x2-1;
	if (y2>clip.y2) y2=clip.y2-1;
	
	long ofs = y*vesa_mode_data->BytesPerScanLine + x*vesa_bytes_per_pixel;
	if (vesa_bytes_per_pixel==2) ofs /= 2;
	if (vesa_bytes_per_pixel==4) ofs /= 4;
	unsigned short *point16 = (unsigned short *)vbuff;
	unsigned long *point32 = (unsigned long *)vbuff;
	
	int turhamuuttuja;
	
	if (vesa_bytes_per_pixel==2)
  {
		if (x==x2)
		{
			for (turhamuuttuja=y;turhamuuttuja<=y2;turhamuuttuja++)
			{
				point16[ofs] = attr;
				ofs += vesa_mode_data->BytesPerScanLine/2;
			}
		} else {
			for (turhamuuttuja=x;turhamuuttuja<=x2;turhamuuttuja++)
			{
				point16[ofs++] = attr;
			}
		}
	}
	
	if (vesa_bytes_per_pixel==4)
	{
		if (x==x2)
		{
			for (turhamuuttuja=y;turhamuuttuja<=y2;turhamuuttuja++)
			{
				point32[ofs] = attr;
				ofs += vesa_mode_data->BytesPerScanLine/4;
			}
		} else {
			for (turhamuuttuja=x;turhamuuttuja<=x2;turhamuuttuja++)
			{
				point32[ofs++] = attr;
			}
		}
	}

	
	if (vesa_bytes_per_pixel==1)
	{
		if (x==x2)
		{
			for (turhamuuttuja=y;turhamuuttuja<=y2;turhamuuttuja++)
			{
				vbuff[ofs] = attr;
				ofs += vesa_mode_data->BytesPerScanLine;
			}
		} else {
			for (turhamuuttuja=x;turhamuuttuja<=x2;turhamuuttuja++)
			{
				vbuff[ofs++] = attr;
			}
		}
	}
}






void draw_box(unsigned char *vbuff, short x, short y, short x2, short y2, long attr, cliptype clip)
{
	short temp;
	if (x2<x) { temp=x; x=x2; x2=temp; }
	if (y2<y) { temp=y; y=y2; y2=temp; }
	
	if (x<clip.x1) x=clip.x1;
	if (y<clip.y1) y=clip.y1;
	if (x2>clip.x2-1) x2=clip.x2-1;
	if (y2>clip.y2-1) y2=clip.y2-1;
	
	long ofs = y*vesa_mode_data->BytesPerScanLine;
  if (vesa_bytes_per_pixel==2) ofs /= 2;
	if (vesa_bytes_per_pixel==4) ofs /= 4;

	unsigned short *point16 = (unsigned short *)vbuff;
	unsigned long *point32 = (unsigned long *)vbuff;
	int turhamuuttuja;
	
	if (vesa_bytes_per_pixel==2)
  {
		for (turhamuuttuja=y;turhamuuttuja<=y2;turhamuuttuja++)
		{
  		for (int xx=x;xx<=x2;xx++) point16[ofs+xx] = attr;
			ofs += vesa_mode_data->BytesPerScanLine/2;
		}
	}
	
	if (vesa_bytes_per_pixel==4)
  {
		for (turhamuuttuja=y;turhamuuttuja<=y2;turhamuuttuja++)
		{
  		for (int xx=x;xx<=x2;xx++) point32[ofs+xx] = attr;
			ofs += vesa_mode_data->BytesPerScanLine/4;
		}
	}

	
	if (vesa_bytes_per_pixel==1)
  {
		for (turhamuuttuja=y;turhamuuttuja<=y2;turhamuuttuja++)
		{
  		for (int xx=x;xx<=x2;xx++) vbuff[ofs+xx] = attr;
			ofs += vesa_mode_data->BytesPerScanLine;
		}
	}
}


void load_sprite(sprite *sp, char *filename, short xsize, short ysize, 
								 short frames, long offset, char *pal, short mode)
{
	sp->frames = frames;
	fstream file;
	file.open(filename, ios::in | ios::binary | ios::nocreate);
	if (!file.good()) return;
	char byte;
	sp->data = (bitmaptype *)malloc(frames*sizeof(bitmaptype));
	bitmapdata trans;
	#if (defined(hicolor) && !defined(pal256))
		trans.r = 0;
		trans.g = 0;
		trans.b = 0;
	#else
		trans = 0;
	#endif
	
	for (short i=0;i<frames;i++)
	{
		sp->data[i].data = (bitmapdata *)malloc(xsize*ysize*sizeof(bitmapdata));
		sp->data[i].xsize = xsize;
		sp->data[i].ysize = ysize;
		sp->data[i].mode = mode;
		sp->data[i].trans = trans;
		for (short j=0;j<xsize*ysize;j++)
		{
			file.read(&byte,1);
			#if (defined(hicolor) && !defined(pal256))
				sp->data[i].data[j].r = pal[byte*3+0]*2;
				sp->data[i].data[j].g = pal[byte*3+1]*2;
				sp->data[i].data[j].b = pal[byte*3+2]*2;
			#else
				sp->data[i].data[j] = byte;
			#endif		
		}
	}
	file.close();
}




void put_sprite(unsigned char *vbuff, long x, long y, 
								sprite *sp, short frame, unsigned char *ttable,
								cliptype clip)
{
	putimg (vbuff, x, y, &sp->data[frame%(sp->frames+1)], ttable, clip);
}
#endif

